/*
 * ActionScript Thread Library
 * 
 * Licensed under the MIT License
 * 
 * Copyright (c) 2010 Leichtgewicht (www.leichtgewicht.at) and
 *                    Spark project (www.libspark.org)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 */
package org.libspark.thread.step {
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.utils.Dictionary;
	import flash.utils.getTimer;

	/**
	 * Processor that executes StepThread instances.
	 * 
	 * <p>
	 *   The execution of StepThread instances requires a global execution model
	 *   that ensures two things:
	 *   <ol>
	 *     <li>That many threads can be executed in parallel</li>
	 *     <li>That threads that required the same resource are not executed in
	 *         parallel</li>
	 *   </ol>
	 * </p>
	 * 
	 * <p>
	 *   This processor implements those requirements. If a thread is started 
	 *   with <code>start</code> its automatically added to a execution list that
	 *   will be processed on every <code>Event.ENTER_FRAME</code>. The method
	 *   <code>lockTo</code> locks a thread to a object and stops it if another
	 *   thread was locked to it first. After a thread was finished all locks
	 *   will be released. But its still possible to unlock it with <code>unlock</code>
	 *   manually. Executing <code>lockTo</code> a second time will unlock it
	 *   all first and eventually allow other threads to continue. This has to
	 *   be done in order to prevent circular thread locks.
	 * </p>
	 * 
	 * <p>
	 *   After every frame, the threads will be adviced to update their status
	 *   this way its not necessary for them to update it on every method call
	 *   (which might be significantly more often per frame). 
	 * </p>
	 * 
	 * @author Martin Heidegger
	 * @version 1.0
	 */
	public class StepProcessor {
		
		/**
		 * Change public access for processor to limit all threads in execution
		 * Useful if 25 milliseconds are still too much.
		 */
		public static var MAX_EXECUTION_TIME_PER_FRAME_IN_MS: int = 25;
		
		// All threads that should be executed in the frame-update-script
		private static var _executingThreads: ThreadExecutionList = new ThreadExecutionList();
		
		// Map of all the threads that have been started
		// StepThread -> Boolean (or null)
		private static const _startedThreads: Dictionary = new Dictionary();
		
		// Lookup for the StepThreads that lock objects
		// * -> StepThread
		private static const _objectLockingThread: Dictionary = new Dictionary();
		
		// Lookup for all the threads which wait for one object.
		// * -> Array( PseudoThread )
		private static const _threadsWaitingForObject: Dictionary = new Dictionary();
		
		// Lookup for all objects a thread locked to
		// StepThread -> Array( * )
		private static const _requiredLockedByThread: Dictionary = new Dictionary();
		
		// Internal eventprovider that sends the Event.ENTER_FRAME trigger
		private static var _eventProvider: Sprite;
		
		// Even though the MAX_EXECUTION_TIME_PER_FRAME_IN_MS is at a defined value
		// we can not make 100% sure that the execution will take exactly so long.
		// If it takes longer, this variable will store how much longer and removes
		// it from the offset of the next frame. This way the frame rate is a lot
		// more stable.
		private static var _formerExecutionOverhead: int;
		
		/**
		 * Adds a thread to the execution list.
		 * 
		 * @param thread StepThread to be started
		 */
		internal static function start( thread: StepThread ): void {
			_executingThreads.add( thread );
			_startedThreads[ thread ] = true;
			if( !_eventProvider ) {
				// Reset the former Execution overhead
				_formerExecutionOverhead = 0;
				_eventProvider = new Sprite();
				// TODO: Implement a proper way to immediatly finish this thread
				// wait one frame at least
				_eventProvider.addEventListener( Event.ENTER_FRAME, processThreads );
			}
		}
		
		/**
		 * Removes the thread from execution list due to external interruption
		 * request.
		 * 
		 * @param thread StepThread to be interrupted
		 */
		internal static function interrupt( thread: StepThread ): void {
			// Clear all locks created by the Thread
			unlock( thread );
			_executingThreads.remove( thread );
			delete _startedThreads[ thread ];
		}
		
		/**
		 * Locks certain objects to a process. This means until next unlock call
		 * no other process that locks to the same objects is allowed to
		 * be executed.
		 * 
		 * <p>If some of the objects have been locked by other processes this
		 * process will be put on hold until all objects are free to be accessed</p>
		 * 
		 * @param thread StepThread that should lock the objects
		 * @param objects List of objects that should be locked
		 * @return true if the locking was successful, false if it was locked by
		 *         some other thread (which sends passed-in thread on hold) or if
		 *         the passed-in thread was not running.
		 */
		internal static function lockTo( thread: StepThread, objects: Array ): Boolean {
			if( objects && thread && _startedThreads[ thread ] ) {
				// Unlock former objects, if some was locked
				unlock( thread );
				// To have all objects available for unlock, pass them in the
				// lookup
				_requiredLockedByThread[ thread ] = objects;
				var wait: Boolean = false;
				var i: int = objects.length;
				while( --i > -1 ) {
					var object: * = objects[ i ];
					// Lock the object == register as locking thread
					if( !_objectLockingThread[ object ] ) {
						_objectLockingThread[ object ] = thread;
					} else {
						// Wait for the object == register as waiting thread
						var waitingThreads: Array = _threadsWaitingForObject[ object ] || ( _threadsWaitingForObject[ object ] = [] );
						// Its possible to assume that the StepThread has not
						// been added before because unlock was called 
						waitingThreads.push( thread );
						wait = true;
					}
				}
				// Its waiting for something, put it on hold
				if( wait ) {
					_executingThreads.remove( thread );
				}
				// If its not waiting its still executing
				return !wait;
			}
			return false;
		}
		
		/**
		 * Unlocks all objects that are locked by a thread.
		 * 
		 * <p>After threads were locked with lock they have to be unlocked at
		 * interrupt or at finish at least. This method can also be called manually
		 * in order to release all locks former taken.</p>
		 * 
		 * @param thread StepThread that should unlock the objects
		 * @param objects List of objects that should be unlocked
		 * @return true if some objects were locked before or if the thread was put
		 *         on hold due to hold, false if no objects were required to be locked
		 *         or if the thread is null
		 */
		internal static function unlock( thread: StepThread ): Boolean {
			var objects: Array = _requiredLockedByThread[ thread ];
			if( thread && objects ) {
				for( var i: int = 0; i < objects.length; ++ i ) {
					var object: * = objects[ i ];
					var waitingThreads: Array = _threadsWaitingForObject[ object ];
					if( waitingThreads ) {
						if( _objectLockingThread[ object ] == thread ) {
							// Take the next thread that is waiting for the
							// object and let it actively lock the object
							var nextThread: StepThread = waitingThreads.shift();
							_objectLockingThread[ object ] = nextThread;
							
							if( waitingThreads.length == 0 ) {
								// None is waiting anymore, trash the list
								delete _threadsWaitingForObject[ object ];
							}
							
							updateExecutionStatusAfterLockChange( nextThread );
						} else if( waitingThreads.length > 1 ) {
							// More threads than this one was waiting for 
							// that object, just splice it
							waitingThreads.splice( waitingThreads.indexOf(thread), 1 );
						} else {
							// Only this thread was waiting, remove the whole
							// waiting list
							delete _threadsWaitingForObject[ object ];
						}
					} else {
						delete _objectLockingThread[ object ];
					}
				}
				// Remove any available locks
				delete _requiredLockedByThread[ thread ];
				// Add the thread as executing again, its not locking anything
				_executingThreads.add( thread );
				return true;
			}
			return false;
		}
		
		/**
		 * Internal helper that checks if all objects a thread was locked to
		 * are finally locked by thread, if thats the case it will be started.
		 * 
		 * @param thread StepThread that should be checked.
		 */
		private static function updateExecutionStatusAfterLockChange( thread: StepThread ): void {
			var objects: Array = _requiredLockedByThread[ thread ];
			var i: int = objects.length;
			while( --i > -1 ) {
				if( _objectLockingThread[ objects[i] ] != thread ) {
					// It has been removed before from execution list (since locks
					// are only being pushed)
					return;
				}
			}
			// No object was not locked by this thread -> add to execution list
			_executingThreads.add( thread );
		}
		
		/**
		 * Processes all threads that are added to the list.
		 * 
		 * @param event ENTER_FRAME event sent by the _eventProvider
		 */
		private static function processThreads( event: Event ): void {
			// Precalculate the max time of all executions that can
			// be done in this frame
			const maxTime: int = getTimer()
			                     + MAX_EXECUTION_TIME_PER_FRAME_IN_MS
			                     - _formerExecutionOverhead;
			
			var current: ThreadExecutionListNode = _executingThreads.nextExecutingThread;
			var i: int = 0;
			try {
				// If _formerExecutionOverhead is bigger than MAX_EXECUTION_TIME_PER_FRAME_IN_MS
				// this statement might even catch the first command
				while( getTimer() < maxTime ) {
					_executingThreads.nextExecutingThread = current.nextExecuting;
					if( current.thread.triggerProcess() ) {
						finalize( current.thread );
					}
					++i;
					
					// circle through the processes, this way each
					// processed will be done in a row.
					current = _executingThreads.nextExecutingThread;
					
					// In case no thread is left ...
					if( !current ) {
						break;
					}
				}
			} catch( e: Error ) {
				finalize( current.thread, e );
			}
			
				
			// Store the current time in order to never get
			// too much out of balance
			_formerExecutionOverhead = ( getTimer() - maxTime )
			                           - MAX_EXECUTION_TIME_PER_FRAME_IN_MS;
			
			if( _formerExecutionOverhead < 0 ) {
				_formerExecutionOverhead = 0;
			}
			
			// TODO: Think of a linear growing system that allows to update
			//       only those which have really been executed.
			
			// Update the state just after each execution
			// Less performance intense
			current = _executingThreads.first;
			try {
				while( current ) {
					_executingThreads.nextThread = current.next;
					current.thread.triggerUpdateState();
					current = _executingThreads.nextThread;
				}
			} catch( e: Error ) {
				// thread.updateState might throw an error
				finalize( current.thread, e );
			}
			
			if( !_executingThreads.first ) {
				// Stop listen to ENTER_FRAME when no thread is left to execute
				_eventProvider.removeEventListener( Event.ENTER_FRAME, processThreads );
				_eventProvider = null;
			}
		}
		
		/**
		 * Finalizes the thread that was started with .start
		 * 
		 * @param thread StepThread that should be finalized
		 * @param error Error that eventually occures which leads the Thread to this
		 *              finalization
		 */
		private static function finalize( thread: StepThread, error: Error = null ) : void {
			unlock( thread );
			_executingThreads.remove( thread );
			try {
				thread.triggerFinish( error );
			} catch( e: Error ) {
				// Global error handling?
				trace( e );
			}
			delete _startedThreads[ thread ];
		}
	}
}
