/*
 * ActionScript Thread Library
 * 
 * Licensed under the MIT License
 * 
 * Copyright (c) 2010 Leichtgewicht (www.leichtgewicht.at) and
 *                    Spark project (www.libspark.org)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 */
package org.libspark.thread.step {
	import org.libspark.thread.Thread;

	import flash.events.Event;
	import flash.events.EventDispatcher;

	/**
	 * StepThread is a abstract class that makes it easy to implement threads
	 * which actively take cpu time to execute without blocking any UI.
	 * 
	 * <p>In order to write a StepThread you have to split up your process into
	 *   small steps. Each step will be executed by StepProcessor as often per
	 *   frame as possible. The amount of things to be done within a thread
	 *   depends strongly on the final purpos. Generally it doesn't make sense
	 *   to have a average processing time per step of less than a millisecond.
	 *   The overhead will be pretty high. Also if the average processing time
	 *   is of more than the frame speed per fps is not very useful since you
	 *   it might be common to create stutterers (not complete shutdowns) in
	 *   interaction with the UI.
	 * </p>
	 * 
	 * <p>Every StepThread that is beeing executed has the possibilty to update
	 *   its status after a the executions of a frame. This way its better performing
	 *   if its possible to execute the thread often per frame.
	 * </p>
	 * 
	 * <p>In order for objects not to be touched by other StepThreads during the
	 *   execution of a thread its possible to lock a thread to objects. After
	 *   the thread was locked to objects it will only be executed once its sure
	 *   no other thread will attempt to start these.
	 *   Of course this is only possible if all StepThreads inform properly about
	 *   the objects that they need.
	 * </p>
	 * 
	 * <p>A example StepThread could look like this:</p>
	 * 
	 * <pre>
	 * package {
	 *     import org.libspark.thread.utils.Progress;
	 *     import org.libspark.thread.step.StepThread;
	 *     
	 *     public class FibonacciThread extends StepThread {
	 *     
	 *     private var _depth: int;
	 *     private var _a: HugeInt;
	 *     private var _b: HugeInt;
	 *     private var _i: int;
	 *     private var _progress: Progress;
	 *     
	 *     public function FibonacciThread( depth: int = 1000 ) {
	 *       if( depth < 2 ) {
	 *         throw new Error( "A depth that is smaller than 2 doesn't need evaluation" );
	 *       }
	 *       _a = new HugeInt( 0 );
	 *       _b = new HugeInt( 1 );
	 *       _progress = new Progress();
	 *       _progress.start( depth-1.0 );
	 *       _depth = depth - 1;
	 *       _i = 0;
	 *     }
	 *     
	 *     public function get result(): String {
	 *       return _b.toString();
	 *     }
	 *     
	 *     override protected function process(): Boolean {
	 *       if( _i < _depth )
	 *       {
	 *         var temp: HugeInt = _a.add( _b );
	 *         _a = _b;
	 *         _b = temp;
	 *         ++_i;
	 *         return false;
	 *       }
	 *       return true;
	 *     }
	 *     
	 *     override protected function updateState(): void {
	 *        _progress.progress( _i );
	 *     }
	 *     
	 *     public function get depth() : int {
	 *       return _depth+1;
	 *     }
	 *   }
	 * }
	 * 
	 * class HugeInt {
	 * 
	 *   private var _parts: Array = new Array();
	 *   
	 *   function HugeInt( init: int ) {
	 *     _parts = [int( init % 10000 ), int( int( init / 10000 ) % 10000 ), int( int( init/100000000 ) % 100000000 ) ];
	 *   }
	 *   
	 *   internal function add( value: HugeInt ): HugeInt {
	 *     var levels: int = value._parts.length;
	 *     var add: int = 0;
	 *     for( var i: int = 0; i< levels; ++i ) {
	 *       var a: int = value._parts[i];
	 *       var b: int = _parts[i];
	 *       var result: int = add;
	 *       if( isNaN( a ) ) {
	 *         result += b;
	 *       } else if( isNaN( b ) ) {
	 *         result += a;
	 *       } else {
	 *         result += a+b;
	 *       }
	 *       _parts[i] = result % 10000;
	 *       add = int( result / 10000 );
	 *     }
	 *     if( add > 0 ) {
	 *       _parts.push( add );
	 *     }
	 *     return this;
	 *   }
	 *   
	 *   internal function toString(): String {
	 *     var result: String = "";
	 *     var i: int = _parts.length;
	 *     while( --i > -1 ) {
	 *       var num: int = _parts[ i ];
	 *       if( num != 0 ) {
	 *         result += num;
	 *         while( --i > -1 ) {
	 *           var no: String = _parts[i];
	 *           while( no.length < 4 ) no = "0"+no;
	 *           result += no;
	 *         }
	 *       }
	 *     }
	 *     return result;
	 *   }
	 * }
	 * </pre>
	 * 
	 * @see StepProcessor
	 * @author Martin Heidegger
	 * @version 1.0
	 */
	public class StepThread extends Thread {
		
		// Internal eventdispatcher that helps to trigger the common
		// Thread events properly
		private const _helpDispatcher: EventDispatcher = new EventDispatcher();
		
		// Internal storage for any error that exited the application.
		private var _error: Error;
		
		/**
		 * Implementation of abstract method of Thread to start the thread.
		 * This method will call the abstract menthod "init"
		 */
		override final protected function run(): void {
			event( _helpDispatcher, Event.COMPLETE, finish );
			StepProcessor.start( this );
			interrupted( removeFromProcessor );
			init();
		}
		
		/**
		 * Template method for extening classes, method will be called
		 * at startup.
		 */
		protected function init(): void {}
		
		/**
		 * Template method to process one step.
		 * 
		 * @param True if the thread is now finished, false if it needs more time
		 */
		protected function process(): Boolean {
			return true;
		}
		
		/**
		 * Template method to update the state of the process.
		 * 
		 * <p>In here you can update the state, will be called once per frame in 
		 * opposite to process which might be called often per frame.</p>
		 */
		protected function updateState(): void {}
		
		/**
		 * Locks certain objects to a process.
		 * 
		 * @return true if the locking was successful, false if it was locked by
		 *         some other thread (which sends passed-in thread on hold) or if
		 *         this instance was not running.
		 * 
		 * @see lockTo
		 * @see StepProcessor.lockTo
		 * @see StepProcessor.unlock
		 */
		protected function lockTo( ...objects: Array ): Boolean {
			return StepProcessor.lockTo( this, objects );
		}
		
		/**
		 * Unlocks all locked objects and continues execution of this thread.
		 * 
		 * @return true if some objects were locked before or if the thread was put
		 *         on hold due to hold, false if no objects were required to be locked
		 *         or if the thread is null
		 * 
		 * @see lockTo
		 * @see StepProcessor.lockTo
		 * @see StepProcessor.unlock
		 */
		protected function unlock(): Boolean {
			return StepProcessor.unlock( this );
		}
		
		/**
		 * Internal method telling the process template triggering
		 * the template method "process".
		 */
		internal function triggerProcess(): Boolean {
			return process();
		}
		
		/**
		 * Internal method telling the updateState template triggering
		 * the template method updateState.
		 */
		internal function triggerUpdateState(): void {
			updateState();
		}
		
		/**
		 * Method to be accessed by StepProcessor. Initializes the finish of the
		 * application by calling Event.COMPLETE to trigger the event handler that
		 * has been added using startard Thread API.
		 * 
		 * @param error Optional error which resulted in this finish.
		 */
		internal function triggerFinish( error: Error = null ): void {
			_error = error;
			_helpDispatcher.dispatchEvent( new Event( Event.COMPLETE ) );
		}
		
		/**
		 * Internal helper that deals as error broadcaster to the system.
		 * 
		 * @param event Event of the system - to be ignored
		 */
		private function finish( event: Event = null ) : void {
			if( _error ) { throw _error; }
		}
		
		/**
		 * Remove this Thread from Processor.
		 * <p>Small util method to handle interrupt event properly.</p>
		 */
		private function removeFromProcessor() : void {
			StepProcessor.interrupt( this);
		}
	}
}
