/*
 * ActionScript Thread Library
 * 
 * Licensed under the MIT License
 * 
 * Copyright (c) 2010 Leichtgewicht (www.leichtgewicht.at) and
 *                    Spark project (www.libspark.org)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 */
package org.libspark.thread.step {
	import flash.utils.Dictionary;

	/**
	 * The Stepthread list is a linkedlist to easy and fast iterate
	 * over all threads
	 * 
	 * @author Martin Heidegger
	 * @private
	 * @version 1.0
	 */
	internal class ThreadExecutionList {
		
		// First node in the list, null if no node added
		internal var first: ThreadExecutionListNode;
		
		// Last node in the list, null if no node added
		internal var last: ThreadExecutionListNode;
		
		// Next node in iteration, stored in list to be
		// working properly even if a thread was added during
		// iteration
		internal var nextThread: ThreadExecutionListNode;
		
		// Next executing thread is the position in the thread iteration.
		// Since the tread iteration might stop at any given thread, some
		// threads would not be processed if others always took all the given
		// processing time. With this ring variable its possible
		// to avoid that effect.
		internal var nextExecutingThread: ThreadExecutionListNode;
		
		// Lookup dictionary to fast access the nodes related
		// to a thread.
		//
		// StepThread -> StepThreadListNode
		internal var lookup: Dictionary = new Dictionary( false );

		/**
		 * Adds a StepThread into the operation list.
		 * 
		 * @param stepThread Thread that should be added to the list
		 * @return true if the Thread was added, false if the thread has been added before
		 */
		internal function add( stepThread: StepThread ): Boolean {
			if( !lookup[stepThread] ) {
				
				// TODO: Pooling of this node might be reasonable, think about
				//       it and add pooling or remove this note.
				var node: ThreadExecutionListNode = new ThreadExecutionListNode();
				
				// Common linked list insert
				node.thread = stepThread;
				if( !first ) {
					first = last = node;
					node.nextExecuting = node;
				} else {
					last.next = node;
					last.nextExecuting = node;
					node.prev = last;
					node.nextExecuting = first;
					last = node;
				}
				
				// If no nextThread is set during iteration, this should
				// become the new next thread
				if( !nextThread ) {
					nextThread = node;
				}
				
				// nextExecuting is for ring-list iteration - this property
				// must exist unless the list is empty
				if( !nextExecutingThread || nextExecutingThread == first ) {
					nextExecutingThread = node;
				}
				
				// Update the lookup
				lookup[ stepThread ] = node;
				
				return true;
			}
			return false;
		}
		
		/**
		 * Removes a StepThread from the operation list.
		 * 
		 * @param stepThread Thread that should be remove from the list
		 * @return true if the Thread was removed, false if the thread has not been added.
		 */
		internal function remove( stepThread: StepThread ): Boolean {
			var node: ThreadExecutionListNode = lookup[ stepThread ];
			if( node ) {
				// Common linked list cleaning
				node.thread = null;
				
				// If this thread is the nextThread, let the next thread be
				// the real nextThread - iteration stability
				if( nextThread == node ) {
					nextThread = node.next;
				}
				
				if( node.nextExecuting == node ) {
					nextExecutingThread = null;
				}
				else
				if( nextExecutingThread == node ) {
					nextExecutingThread = node.nextExecuting;
				}
				
				if( node.prev ) {
					if( node.next ) {
						node.next.prev = node.prev;
						node.prev.next = node.next;
						node.prev.nextExecuting = node.next;
					} else {
						last = node.prev;
						last.nextExecuting = first;
						last.next = null;
					}
				} else if( node.next ) {
					first = node.next;
					first.prev = null;
					last.nextExecuting = first;
				} else {
					first = null;
					last = null;
				}
				
				node.nextExecuting = null;
				
				delete lookup[ stepThread ];
				
				return true;
			}
			return false;
		}
	}
}
